/*
 * Pickwick Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * A kqueue-based, multiplexing chat server, using xor obfuscation.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.14"
#define PIDFILE "/var/run/pickwick.pid"

#define XOR_CHAR 129

/*
 * Maximum size of connection output queues we use to cope with
 * network congestion.
 */

#define MAX_QUEUE 500

/*
 * Maximum size of chunks of data read and written from/to client.  We
 * multiplex by only processing this much data each time we service a
 * client connection.
 */

#define IOBUFSIZE 1024

/*
 * Maximum length of user names.  This value is 25 x 5, so that if the
 * user is using UTF-8, they may have a 25 character name.
 */

#define MAX_NAME 125

/*
 * KQueue Input and output queues.  We add 4 events for each accept()ed
 * descriptor to the input queue.  We may add up to 4 events for
 * already-existing connections.
 */

struct kevent *inqueue = NULL;
struct kevent *outqueue = NULL;

/*
 * We drop connections which go idle this number of seconds.
 */

#define TIMEOUT 60 * 20

extern char *optarg;

int in = 0, out = 0, fd = -1, logging = 0, testing = 0,
    backlog = 1000, closed = 0, conn_count = 0, max_conn = 2,
    qlen = 6, inq = 24;

char *interface = "", *port = "7331", *grp = "nobody", *user = "nobody",
     *mask = "<%s is online>\n";

/*
 * Declared as character arrays, so we may legally overwrite them.
 */

char err_msg[]   = "\nMaximum number of clients already connected.  Sorry.\n",
     wel_msg[]   = "\n----------------------------\n"
                   "  Welcome to Pickwick " VERSION "\n"
                   "----------------------------\n\n",

     no_msg[]    = "<no one else is online>\n",
     chs_msg[]   = "Please choose a user name: ",
     act_msg[]   = "<inactivity timeout>\n",

     colon[]     = ": ",
     newline[]   = "\n",
     ellipses[]  = "...\n",

     who_cmd[]   = "who",
     exit_cmd[]  = "exit",
     quit_cmd[]  = "quit",
     help_cmd[]  = "help",
     bad_cmd[]   = "<unrecognized command>\n",

     help_msg[]  = "<available commands:>\n"
                   "<:who  - show who is online>\n"
                   "<:exit - close connection>\n"
                   "<:quit - close connection>\n"
                   "<:help - display this list>\n",

     taken_msg[] = "That name is taken.  Choose another: ",
     long_msg[]  = "That name is too long.  Choose another: ";

struct passwd *passwd;
struct group *group;

/*
 * Queue of buffered out-going data for each connection.
 */

struct queue {
   int total, count;
   char buffer[ IOBUFSIZE ];
   struct queue *next;
};

/*
 * Connection control block.  name has extra space for \n\0.
 */

struct ccb {
   int sock, count, total, qlen, named, complete;
   char name[ MAX_NAME + 2 ];
   struct queue *first, *last;
} *conns = NULL;

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "pickwick: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

void delete_timer( struct ccb *item, int n, int out )
{
   struct kevent *kev;

   if ( in >= inq )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = 0;
   kev->flags = EV_DELETE;
   kev->udata = item;

   while( ++n < out )
      if ( outqueue[ n ].ident == item->sock &&
           outqueue[ n ].filter == EVFILT_TIMER )
         outqueue[ n ].ident = -1;
}

void set_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= inq )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = TIMEOUT * 1000;
   kev->flags = EV_ADD;
   kev->udata = item;
}

void ev_set( int desc, short filter, u_short flags, struct ccb *item )
{
   struct kevent *kev;

   if ( in >= inq )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = desc;
   kev->filter = filter;
   kev->fflags = 0;
   kev->flags = flags;
   kev->udata = item;
}

void non_blocking( int desc )
{
   int flags, unblocked;

   if (( flags = fcntl( desc, F_GETFL, 0 )) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }

   unblocked = flags & O_NONBLOCK;

   if ( ! unblocked && fcntl( desc, F_SETFL, flags | O_NONBLOCK ) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }
}

void set_options( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xm:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'g':
            grp = ( optarg == NULL ? "nobody" : optarg );
            break;

         case 'i':
            interface = ( optarg == NULL ? "" : optarg );
            break;

         case 'm':
            max_conn = strtol( optarg, NULL, 10 );
            break;

         case 'p':
            port = ( optarg == NULL ? "7" : optarg );
            break;

         case 'u':
            user = ( optarg == NULL ? "nobody" : optarg );
            break;

         case 'x':
            testing = 1;
      }

   if ( max_conn <= 0 )
   {
      fprintf( stderr, "pickwick: max_conn <= 0: %d\n", max_conn );
      exit( 1 );
   }

   qlen = max_conn * 3;
   inq = qlen * 4;

   if (( conns = memory( sizeof( struct ccb ) * max_conn )) == NULL )
      exit( 1 );

   if (( outqueue = memory( sizeof( struct kevent ) * qlen )) == NULL )
      exit( 1 );

   if (( inqueue = memory( sizeof( struct kevent ) * inq )) == NULL )
      exit( 1 );

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "pickwick: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "pickwick: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }
}

void become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "pickwick: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if ( ! testing )
   {
      if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC )) < 0 )
         syslog( LOG_WARNING, "open(): %m" );
      else
      {
         char buffer[ 16 ];

         snprintf( buffer, sizeof( buffer ), "%d", getpid() );
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   /*
    * We don't create any files or fork children which create files,
    * but if we did, this will make those files not world-readable by
    * default.
    */

   umask( 2 );
}

void change_identity()
{
   if ( setgid( group->gr_gid ) < 0 )
      syslog( LOG_ERR, "setgid(): %m" );

   if ( setuid( passwd->pw_uid ) < 0 )
      syslog( LOG_ERR, "setuid(): %m" );
}

void start_listening()
{
   struct addrinfo hints, *res;
   int result;

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can restart immediately.
    */

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   non_blocking( fd );
}

/*
 * We do not encode the XOR_CHAR in the data stream, to avoid embedding
 * zeros in data strings.  This is possible with UTF-8 input.
 */

void encode_data( char *buffer )
{
   while( *buffer )
      if (( unsigned char)*buffer != XOR_CHAR )
         *buffer++ ^= XOR_CHAR;
}

void enqueue_data( struct ccb *item, char *data, int len )
{
   struct queue *q;

   if (( q = memory( sizeof( struct queue ))) == NULL )
      return;

   if ( ++item->qlen > MAX_QUEUE )
      return;

   if ( item->first == NULL )
   {
      item->first = item->last = q;
      ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
   }
   else
   {
      item->last->next = q;
      item->last = q;
   }

   item->last->next = NULL;

   item->last->count = 0;
   item->last->total = ( len < 0 ? strlen( data ) : len );

   bcopy( data, item->last->buffer, item->last->total );
}

void add_conn( int new )
{
   int i;
   struct ccb *ptr = NULL;

   /*
    * The socket is still blocking at this point.
    */

   if ( conn_count == max_conn )
   {
      write( new, err_msg, strlen( err_msg ));
      close( new );
      return;
   }

   for( ptr = conns, i = 0; i < max_conn; ++ptr, ++i )
      if ( ! ptr->sock )
         break;

   ++conn_count;
   non_blocking( new );

   ptr->sock = new;
   *ptr->name = '\0';

   ptr->qlen = ptr->total = ptr->count = ptr->named = 0;
   ptr->first = ptr->last = NULL;

   ptr->complete = 1;

   ev_set( ptr->sock, EVFILT_READ, EV_ADD | EV_DISABLE, ptr );
   ev_set( ptr->sock, EVFILT_WRITE, EV_ADD | EV_DISABLE, ptr );
   set_timer( ptr );

   enqueue_data( ptr, wel_msg, -1 );
   enqueue_data( ptr, chs_msg, -1 );
}

void remove_conn( struct ccb *item )
{
   char buffer[ IOBUFSIZE ];
   struct ccb *ptr;
   int i;

   /*
    * remove_events() will be called when we return to process_clients(), because
    * of closed being set below.  Any events pending for this connection in the
    * output queue, including timer events, will be nullified there.  So we
    * can pass zeros to delete_timer() here.
    */

   delete_timer( item, 0, 0 );

   --conn_count;
   closed = item->sock;
   close( item->sock );

   if ( item->named )
   {
      encode_data( item->name );

      snprintf( buffer, sizeof( buffer ),
                "<%s has left the conversation>\n", item->name );

      encode_data( buffer );
      encode_data( item->name );

      for( ptr = conns, i = 0; i < max_conn; ++ptr, ++i )
         if ( ptr != item && ptr->named )
            enqueue_data( ptr, buffer, -1 );
   }

   while( item->first != NULL )
   {
      item->last = item->first->next;
      free( item->first );
      item->first = item->last;
   }

   bzero( item, sizeof( struct ccb ));
}

void accept_connection()
{
   struct sockaddr_storage addr;
   socklen_t len;
   int conn;

   len = sizeof( struct sockaddr_storage );
   conn = accept( fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         syslog( LOG_ERR, "accept(): %m" );
   }
   else
      add_conn( conn );
}

void get_name( struct ccb *item )
{
   int i, count;
   char *ptr;
   struct ccb *cptr;
   static char buffer[ IOBUFSIZE ];

   if (( count = read( item->sock, &item->name[ item->total ],
                       ( MAX_NAME + 1 ) - item->count )) <= 0 )
   {
      remove_conn( item );
      return;
   }

   item->total += count;
   item->count -= count;
   item->name[ item->total ] = '\0';

   if (( ptr = strchr( item->name, *newline )) == NULL )
   {
      if ( item->total > MAX_NAME )
      {
         while( read( item->sock, buffer, IOBUFSIZE ) > 0 )
            ;

         *item->name = '\0';
         item->total = item->count = 0;
         enqueue_data( item, long_msg, -1 );
      }

      return;
   }

   *ptr = '\0';
   item->total = item->count = 0;

   /*
    * We will never get a blank line if the client is weller, for that
    * program interprets blank lines as toggles of the screenlock, and
    * does not pass them on.
    */

   if ( ! *item->name )
   {
      enqueue_data( item, chs_msg, -1 );
      return;
   }

   for( cptr = conns, i = 0; i < max_conn; ++cptr, ++i )
      if ( cptr->named && ! strcmp( cptr->name, item->name ))
      {
         *item->name = '\0';
         enqueue_data( item, taken_msg, -1 );
         return;
      }

   for( cptr = conns, count = 0, i = 0; i < max_conn; ++cptr, ++i )
      if ( cptr->named )
      {
         ++count;

         encode_data( cptr->name );
         snprintf( buffer, sizeof( buffer ), mask, cptr->name );
         encode_data( buffer );
         encode_data( cptr->name );

         enqueue_data( item, buffer, -1 );
      }

   if ( ! count )
      enqueue_data( item, no_msg, -1 );

   item->named = 1;
   encode_data( item->name );
   snprintf( buffer, sizeof( buffer ),
             "<%s has joined the conversation>\n", item->name );
   encode_data( buffer );
   encode_data( item->name );

   for( cptr = conns, i = 0; i < max_conn; ++cptr, ++i )
      if ( cptr->named && cptr != item )
         enqueue_data( cptr, buffer, -1 );

   enqueue_data( item, help_msg, -1 );
}

void show_who( struct ccb *item )
{
   static char buffer[ IOBUFSIZE ];
   struct ccb *cptr;
   int count, i;

   for( cptr = conns, count = 0, i = 0; i < max_conn; ++cptr, ++i )
      if ( cptr->named && cptr != item )
      {
         ++count;

         encode_data( cptr->name );
         snprintf( buffer, sizeof( buffer ), mask, cptr->name );
         encode_data( buffer );
         encode_data( cptr->name );

         enqueue_data( item, buffer, -1 );
      }

   if ( ! count )
      enqueue_data( item, no_msg, -1 );
}

void command( struct ccb *item, char *cmd, int n )
{
   if ( !strncasecmp( who_cmd, cmd, n ))
      show_who( item );
   else if ( !strncasecmp( exit_cmd, cmd, n ) || !strncasecmp( quit_cmd, cmd, n ))
      remove_conn( item );
   else if ( !strncasecmp( help_cmd, cmd, n ))
      enqueue_data( item, help_msg, -1 );
   else
      enqueue_data( item, bad_cmd, -1 );
}

void transfer_out( struct ccb *item )
{
   int count;
   struct queue *q;

   if ( item->first == NULL )
      return;

   if (( count = write( item->sock, &item->first->buffer[ item->first->count ],
                        item->first->total )) < 0 )
   {
      if ( errno == EWOULDBLOCK )
         return;

      syslog( LOG_ERR, "write(): %m" );
      remove_conn( item );
   }

   item->first->total -= count;
   item->first->count += count;

   if ( ! item->first->total )
   {
      q = item->first;
      item->first = item->first->next;
      free( q );
      --item->qlen;

      if ( item->first == NULL )
      {
         ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );

         if ( ! item->named )
            ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );
       }
   }
}

void transfer_in( struct ccb *item )
{
   int i, count, complete;
   struct ccb *ptr;
   static char buffer[ IOBUFSIZE ];

   snprintf( buffer, sizeof( buffer ), "%s%s", item->name, colon );
   i = strlen( buffer );

   if (( count = read( item->sock, &buffer[ i ], IOBUFSIZE - i )) <= 0 )
   {
      remove_conn( item );
      return;
   }

   count += i;
   complete = item->complete;
   
   if (( item->complete = ( buffer[ count - 1 ] == *newline )) &&
        complete &&
        buffer[ i ] == *colon )
      command( item, &buffer[ i + 1 ], count - i - 2 );
   else
      for( ptr = conns, i = 0; i < max_conn; ++ptr, ++i )
      {
         if ( ptr->named )
         {
            enqueue_data( ptr, buffer, count );
      
            if ( ! item->complete )
               enqueue_data( ptr, ellipses, 4 );
         }
      }
}
   
/*
 * If we have dropped a connection, we traverse the output kqueue
 * and mark any events for that connection as invalid, by setting
 * the socket descriptor to -1.
 */

void remove_events( int n )
{
   while( ++n < out )
      if ( outqueue[ n ].ident == closed )
         outqueue[ n ].ident = -1;
}

/*
 * The listening socket must be non-blocking as connections may be dropped
 * between the time the kernel queues them and the time we accept them.

 * We must delete a timer before resetting it, to clear out timer events
 * which may have occurred while we were processing the outqueue.

 * We only accept one new connection for every traversal of
 * the output queue.
 */

void process_clients()
{
   int kq, n;

   if (( kq = kqueue()) < 0 )
   {
      syslog( LOG_ERR, "kqueue(): %m" );
      exit( 1 );
   }

   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, NULL );

   for( ; ; )
   {
      out = kevent( kq, inqueue, in, outqueue, qlen, NULL );
      in = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         struct ccb *item;

         if ( outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         closed = 0;
         item = ( struct ccb *)outqueue[ n ].udata;

         if ( outqueue[ n ].ident == fd )
            accept_connection();
         else if ( outqueue[ n ].filter == EVFILT_TIMER )
         {
            write( item->sock, act_msg, strlen( act_msg ));
            remove_conn( item );
         }
         else if ( outqueue[ n ].filter == EVFILT_READ )
         {
            if ( item->named )
               transfer_in( item );
            else
               get_name( item );

            if ( !closed )
            {
               delete_timer( item, n, out );
               set_timer( item );
            }
         }
         else
            transfer_out( item );

         if ( closed )
            remove_events( n );
      }
   }
}

int main( int argc, char **argv )
{
   int i;
   char *strs[ 17 ], **ptr;

   set_options( argc, argv );
   signal( SIGPIPE, SIG_IGN );

   openlog( "pickwick", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();
   change_identity();

   for( i = 0; i <= max_conn; ++i )
      conns[ i ].sock = 0;

   strs[ 0 ] = err_msg;
   strs[ 1 ] = wel_msg;
   strs[ 2 ] = no_msg;
   strs[ 3 ] = chs_msg;
   strs[ 4 ] = act_msg;
   strs[ 5 ] = colon;
   strs[ 6 ] = taken_msg;
   strs[ 7 ] = long_msg;
   strs[ 8 ] = newline;
   strs[ 9 ] = ellipses;
   strs[ 10 ] = who_cmd;
   strs[ 11 ] = exit_cmd;
   strs[ 12 ] = quit_cmd;
   strs[ 13 ] = bad_cmd;
   strs[ 14 ] = help_cmd;
   strs[ 15 ] = help_msg;
   strs[ 16 ] = NULL;

   for( ptr = strs; *ptr != NULL; ++ptr )
      encode_data( *ptr );

   process_clients();

   return 0;
}
